__name__ = "orix"
__version__ = "0.8.1"
__author__ = "orix developers"
__author_email__ = "pyxem.team@gmail.com"
__description__ = "orix is an open-source python library for handling crystal orientation mapping data."
# Initial committer first, then sorted by line contributions
__credits__ = [
    "Ben Martineau",
    "Håkon Wiik Ånes",
    "Phillip Crout",
    "Duncan Johnstone",
    "Paddy Harrison",
    "Niels Cautaerts",
    "Simon Høgås",
]
